@@ -33,25 +33,9 @@ extern struct itemplate **itable[];
 #define SEG_NODISP 64
 #define SEG_SIGNED 128
 
-static int whichreg(long regflags, int regval) {
-    static int reg32[] = {
-	R_EAX, R_ECX, R_EDX, R_EBX, R_ESP, R_EBP, R_ESI, R_EDI };
-    static int reg16[] = {
-	R_AX, R_CX, R_DX, R_BX, R_SP, R_BP, R_SI, R_DI };
-    static int reg8[] = {
-	R_AL, R_CL, R_DL, R_BL, R_AH, R_CH, R_DH, R_BH };
-    static int sreg[] = {
-	R_ES, R_CS, R_SS, R_DS, R_FS, R_GS, 0, 0 };
-    static int creg[] = {
-	R_CR0, 0, R_CR2, R_CR3, R_CR4, 0, 0, 0 };
-    static int dreg[] = {
-	R_DR0, R_DR1, R_DR2, R_DR3, 0, 0, R_DR6, R_DR7 };
-    static int treg[] = {
-	0, 0, 0, R_TR3, R_TR4, R_TR5, R_TR6, R_TR7 };
-    static int fpureg[] = {
-	R_ST0, R_ST1, R_ST2, R_ST3, R_ST4, R_ST5, R_ST6, R_ST7 };
-    static int mmxreg[] = {
-	R_MM0, R_MM1, R_MM2, R_MM3, R_MM4, R_MM5, R_MM6, R_MM7 };
+static int whichreg(long regflags, int regval) 
+{
+#include "regdis.c"
 
     if (!(REG_AL & ~regflags))
 	return R_AL;
@@ -67,10 +51,21 @@ static int whichreg(long regflags, int regval) {
 	return R_CX;
     if (!(REG_ECX & ~regflags))
 	return R_ECX;
-    if (!(REG_CR4 & ~regflags))
-	return R_CR4;
     if (!(FPU0 & ~regflags))
 	return R_ST0;
+    if (!(REG_CS & ~regflags))
+        return (regval == 1) ? R_CS : 0;
+    if (!(REG_DESS & ~regflags))
+	return (regval == 0 || regval == 2 || regval == 3 ? sreg[regval] : 0);
+    if (!(REG_FSGS & ~regflags))
+	return (regval == 4 || regval == 5 ? sreg[regval] : 0);
+    if (!(REG_SEG67 & ~regflags))
+	return (regval == 6 || regval == 7 ? sreg[regval] : 0);
+
+    /* All the entries below look up regval in an 8-entry array */
+    if (regval < 0 || regval > 7)
+	return 0;
+
     if (!((REGMEM|BITS8) & ~regflags))
 	return reg8[regval];
     if (!((REGMEM|BITS16) & ~regflags))
@@ -89,13 +84,17 @@ static int whichreg(long regflags, int regval) {
 	return fpureg[regval];
     if (!(MMXREG & ~regflags))
 	return mmxreg[regval];
+    if (!(XMMREG & ~regflags))
+	return xmmreg[regval];
+
     return 0;
 }
 
-static char *whichcond(int condval) {
+static const char *whichcond(int condval) 
+{
     static int conds[] = {
-	C_O, C_NO, C_B, C_AE, C_E, C_NE, C_BE, C_A,
-	C_S, C_NS, C_PE, C_PO, C_L, C_GE, C_LE, C_G
+	C_O, C_NO, C_C, C_NC, C_Z, C_NZ, C_NA, C_A,
+	C_S, C_NS, C_PE, C_PO, C_L, C_NL, C_NG, C_G
     };
     return conditions[conds[condval]];
 }
@@ -104,7 +103,8 @@ static char *whichcond(int condval) {
  * Process an effective address (ModRM) specification.
  */
 static unsigned char *do_ea (unsigned char *data, int modrm, int asize,
-			     int segsize, operand *op) {
+			     int segsize, operand *op) 
+{
     int mod, rm, scale, index, base;
 
     mod = (modrm >> 6) & 03;
@@ -239,15 +239,24 @@ static unsigned char *do_ea (unsigned char *data, int modrm, int asize,
 }
 
 /*
- * Determine whether the code string in r corresponds to the data
+ * Determine whether the instruction template in t corresponds to the data
  * stream in data. Return the number of bytes matched if so.
  */
-static int matches (unsigned char *r, unsigned char *data, int asize,
-		    int osize, int segsize, insn *ins) {
-    unsigned char *origdata = data;
-    int a_used = FALSE, o_used = FALSE;
+static int matches (struct itemplate *t, unsigned char *data, int asize,
+		    int osize, int segsize, int rep, insn *ins) 
+{
+    unsigned char * r = (unsigned char *)(t->code);
+    unsigned char * origdata = data;
+    int           a_used = FALSE, o_used = FALSE;
+    int           drep = 0;
 
-    while (*r) {
+    if ( rep == 0xF2 )
+      drep = P_REPNE;
+    else if ( rep == 0xF3 )
+      drep = P_REP;
+
+    while (*r) 
+    {
 	int c = *r++;
 	if (c >= 01 && c <= 03) {
 	    while (c--)
@@ -312,7 +321,7 @@ static int matches (unsigned char *r, unsigned char *data, int asize,
 	if (c >= 034 && c <= 036) {
 	    ins->oprs[c-034].offset = *data++;
 	    ins->oprs[c-034].offset |= (*data++ << 8);
-	    if (asize == 32) {
+	    if (osize == 32) {
 		ins->oprs[c-034].offset |= (((long) *data++) << 16);
 		ins->oprs[c-034].offset |= (((long) *data++) << 24);
 	    }
@@ -325,6 +334,16 @@ static int matches (unsigned char *r, unsigned char *data, int asize,
 	    ins->oprs[c-040].offset |= (((long) *data++) << 16);
 	    ins->oprs[c-040].offset |= (((long) *data++) << 24);
 	}
+	if (c >= 044 && c <= 046) {
+	    ins->oprs[c-044].offset = *data++;
+	    ins->oprs[c-044].offset |= (*data++ << 8);
+	    if (asize == 32) {
+		ins->oprs[c-044].offset |= (((long) *data++) << 16);
+		ins->oprs[c-044].offset |= (((long) *data++) << 24);
+	    }
+	    if (segsize != asize)
+		ins->oprs[c-044].addr_size = asize;
+	}
 	if (c >= 050 && c <= 052) {
 	    ins->oprs[c-050].offset = (signed char) *data++;
 	    ins->oprs[c-050].segment |= SEG_RELATIVE;
@@ -338,15 +357,18 @@ static int matches (unsigned char *r, unsigned char *data, int asize,
 	if (c >= 064 && c <= 066) {
 	    ins->oprs[c-064].offset = *data++;
 	    ins->oprs[c-064].offset |= (*data++ << 8);
-	    if (asize == 32) {
+	    if (osize == 32) {
 		ins->oprs[c-064].offset |= (((long) *data++) << 16);
 		ins->oprs[c-064].offset |= (((long) *data++) << 24);
 		ins->oprs[c-064].segment |= SEG_32BIT;
 	    } else
 		ins->oprs[c-064].segment &= ~SEG_32BIT;
 	    ins->oprs[c-064].segment |= SEG_RELATIVE;
-	    if (segsize != asize)
-		ins->oprs[c-064].addr_size = asize;
+	    if (segsize != osize) {
+	        ins->oprs[c-064].type =
+		    (ins->oprs[c-064].type & NON_SIZE)
+		    | ((osize == 16) ? BITS16 : BITS32);
+	    }
 	}
 	if (c >= 070 && c <= 072) {
 	    ins->oprs[c-070].offset = *data++;
@@ -355,13 +377,23 @@ static int matches (unsigned char *r, unsigned char *data, int asize,
 	    ins->oprs[c-070].offset |= (((long) *data++) << 24);
 	    ins->oprs[c-070].segment |= SEG_32BIT | SEG_RELATIVE;
 	}
-	if (c >= 0100 && c <= 0177) {
+	if (c >= 0100 && c < 0130) {
 	    int modrm = *data++;
 	    ins->oprs[c & 07].basereg = (modrm >> 3) & 07;
 	    ins->oprs[c & 07].segment |= SEG_RMREG;
 	    data = do_ea (data, modrm, asize, segsize,
 			  &ins->oprs[(c >> 3) & 07]);
 	}
+	if (c >= 0130 && c <= 0132) {
+	    ins->oprs[c-0130].offset = *data++;
+	    ins->oprs[c-0130].offset |= (*data++ << 8);
+	}
+	if (c >= 0140 && c <= 0142) {
+	    ins->oprs[c-0140].offset = *data++;
+	    ins->oprs[c-0140].offset |= (*data++ << 8);
+	    ins->oprs[c-0140].offset |= (((long) *data++) << 16);
+	    ins->oprs[c-0140].offset |= (((long) *data++) << 24);
+	}
 	if (c >= 0200 && c <= 0277) {
 	    int modrm = *data++;
 	    if (((modrm >> 3) & 07) != (c & 07))
@@ -419,12 +451,27 @@ static int matches (unsigned char *r, unsigned char *data, int asize,
 	    else
 		ins->condition = d - t;
 	}
+	if (c == 0331) {
+	    if ( rep )
+	        return FALSE;
+	}
+	if (c == 0332) {
+	    if (drep == P_REP)
+	        drep = P_REPE;
+	}
+	if (c == 0333) {
+	    if ( rep != 0xF3 )
+	        return FALSE;
+	    drep = 0;
+	}
     }
 
     /*
-     * Check for unused a/o prefixes.
+     * Check for unused rep or a/o prefixes.
      */
     ins->nprefix = 0;
+    if (drep)
+        ins->prefixes[ins->nprefix++] = drep;
     if (!a_used && asize != segsize)
 	ins->prefixes[ins->nprefix++] = (asize == 16 ? P_A16 : P_A32);
     if (!o_used && osize != segsize)
@@ -434,14 +481,16 @@ static int matches (unsigned char *r, unsigned char *data, int asize,
 }
 
 long disasm (unsigned char *data, char *output, int segsize, long offset,
-	     int autosync) {
-    struct itemplate **p;
-    int length = 0;
+	     int autosync, unsigned long prefer)
+{
+    struct itemplate **p, **best_p;
+    int length, best_length = 0;
     char *segover;
     int rep, lock, asize, osize, i, slen, colon;
     unsigned char *origdata;
     int works;
-    insn ins;
+    insn tmp_ins, ins;
+    unsigned long goodness, best;
 
     /*
      * Scan for prefixes.
@@ -473,51 +522,73 @@ long disasm (unsigned char *data, char *output, int segsize, long offset,
 	    break;
     }
 
-    ins.oprs[0].segment = ins.oprs[1].segment = ins.oprs[2].segment =
-    ins.oprs[0].addr_size = ins.oprs[1].addr_size = ins.oprs[2].addr_size =
-	(segsize == 16 ? 0 : SEG_32BIT);
-    ins.condition = -1;
-    works = TRUE;
-    for (p = itable[*data]; *p; p++)
-	if ( (length = matches((unsigned char *)((*p)->code), data,
-			       asize, osize, segsize, &ins)) ) {
-	    works = TRUE;
-	    /*
-	     * Final check to make sure the types of r/m match up.
-	     */
-	    for (i = 0; i < (*p)->operands; i++)
-		if (((ins.oprs[i].segment & SEG_RMREG) &&
-		     !(MEMORY & ~(*p)->opd[i])) ||
-		    (!(ins.oprs[i].segment & SEG_RMREG) &&
-		     !(REGNORM & ~(*p)->opd[i]) &&
-		     !((*p)->opd[i] & REG_SMASK)))
-		    works = FALSE;
-	    if (works)
-		break;
+    tmp_ins.oprs[0].segment = tmp_ins.oprs[1].segment =
+    tmp_ins.oprs[2].segment =
+    tmp_ins.oprs[0].addr_size = tmp_ins.oprs[1].addr_size =
+      tmp_ins.oprs[2].addr_size = (segsize == 16 ? 0 : SEG_32BIT);
+    tmp_ins.condition = -1;
+    best = ~0UL;		/* Worst possible */
+    best_p = NULL;
+    for (p = itable[*data]; *p; p++) {
+      if ( (length = matches(*p, data, asize, osize,
+			     segsize, rep, &tmp_ins)) ) {
+	works = TRUE;
+	/*
+	 * Final check to make sure the types of r/m match up.
+	 */
+	for (i = 0; i < (*p)->operands; i++) {
+	  if (
+	      /* If it's a mem-only EA but we have a register, die. */
+	      ((tmp_ins.oprs[i].segment & SEG_RMREG) &&
+	       !(MEMORY & ~(*p)->opd[i])) ||
+	      
+	      /* If it's a reg-only EA but we have a memory ref, die. */
+	      (!(tmp_ins.oprs[i].segment & SEG_RMREG) &&
+	       !(REGNORM & ~(*p)->opd[i]) &&
+	       !((*p)->opd[i] & REG_SMASK)) ||
+	      
+	      /* Register type mismatch (eg FS vs REG_DESS): die. */
+	      ((((*p)->opd[i] & (REGISTER | FPUREG)) ||
+		(tmp_ins.oprs[i].segment & SEG_RMREG)) &&
+	       !whichreg ((*p)->opd[i], tmp_ins.oprs[i].basereg))) {
+	    works = FALSE;
+	    break;
+	  }
+	}  
+	
+	if (works) {
+	  goodness = ((*p)->flags & IF_PFMASK) ^ prefer;
+	  if ( goodness < best ) {
+	    /* This is the best one found so far */
+	    best        = goodness;
+	    best_p      = p;
+	    best_length = length;
+	    ins         = tmp_ins;
+	  }
 	}
-    if (!length || !works)
+      }
+    }
+
+    if (!best_p)
 	return 0;		       /* no instruction was matched */
 
+    /* Pick the best match */
+    p      = best_p;
+    length = best_length;
+
     slen = 0;
 
-    if (rep) {
-	slen += sprintf(output+slen, "rep%s ",
-			(rep == 0xF2 ? "ne" :
-			 (*p)->opcode == I_CMPSB ||
-			 (*p)->opcode == I_CMPSW ||
-			 (*p)->opcode == I_CMPSD ||
-			 (*p)->opcode == I_SCASB ||
-			 (*p)->opcode == I_SCASW ||
-			 (*p)->opcode == I_SCASD ? "e" : ""));
-    }
     if (lock)
 	slen += sprintf(output+slen, "lock ");
     for (i = 0; i < ins.nprefix; i++)
 	switch (ins.prefixes[i]) {
-	  case P_A16: slen += sprintf(output+slen, "a16 "); break;
-	  case P_A32: slen += sprintf(output+slen, "a32 "); break;
-	  case P_O16: slen += sprintf(output+slen, "o16 "); break;
-	  case P_O32: slen += sprintf(output+slen, "o32 "); break;
+	  case P_REP:   slen += sprintf(output+slen, "rep "); break;
+	  case P_REPE:  slen += sprintf(output+slen, "repe "); break;
+	  case P_REPNE: slen += sprintf(output+slen, "repne "); break;
+	  case P_A16:   slen += sprintf(output+slen, "a16 "); break;
+	  case P_A32:   slen += sprintf(output+slen, "a32 "); break;
+	  case P_O16:   slen += sprintf(output+slen, "o16 "); break;
+	  case P_O32:   slen += sprintf(output+slen, "o32 "); break;
 	}
 
     for (i = 0; i < elements(ico); i++)
@@ -553,11 +624,14 @@ long disasm (unsigned char *data, char *output, int segsize, long offset,
 	    colon = FALSE;
 
 	if (((*p)->opd[i] & (REGISTER | FPUREG)) ||
-	    (ins.oprs[i].segment & SEG_RMREG)) {
+	    (ins.oprs[i].segment & SEG_RMREG)) 
+	{
 	    ins.oprs[i].basereg = whichreg ((*p)->opd[i],
 					    ins.oprs[i].basereg);
+	    if ( (*p)->opd[i] & TO )
+		slen += sprintf(output+slen, "to ");
 	    slen += sprintf(output+slen, "%s",
-			    reg_names[ins.oprs[i].basereg]);
+			    reg_names[ins.oprs[i].basereg-EXPR_REG_START]);
 	} else if (!(UNITY & ~(*p)->opd[i])) {
 	    output[slen++] = '1';
 	} else if ( (*p)->opd[i] & IMMEDIATE ) {
@@ -615,14 +689,16 @@ long disasm (unsigned char *data, char *output, int segsize, long offset,
 	    }
 	    if (ins.oprs[i].basereg != -1) {
 		slen += sprintf(output+slen, "%s",
-				reg_names[ins.oprs[i].basereg]);
+				reg_names[(ins.oprs[i].basereg -
+					   EXPR_REG_START)]);
 		started = TRUE;
 	    }
 	    if (ins.oprs[i].indexreg != -1) {
 		if (started)
 		    output[slen++] = '+';
 		slen += sprintf(output+slen, "%s",
-				reg_names[ins.oprs[i].indexreg]);
+				reg_names[(ins.oprs[i].indexreg -
+					   EXPR_REG_START)]);
 		if (ins.oprs[i].scale > 1)
 		    slen += sprintf(output+slen, "*%d", ins.oprs[i].scale);
 		started = TRUE;
@@ -661,7 +737,8 @@ long disasm (unsigned char *data, char *output, int segsize, long offset,
     return length;
 }
 
-long eatbyte (unsigned char *data, char *output) {
+long eatbyte (unsigned char *data, char *output) 
+{
     sprintf(output, "db 0x%02X", *data);
     return 1;
 }
