@@ -11,9 +11,8 @@
 #ifndef NASM_NASM_H
 #define NASM_NASM_H
 
-#define NASM_MAJOR_VER 0
-#define NASM_MINOR_VER 97
-#define NASM_VER "0.97"
+#include <stdio.h>
+#include "version.h"		       /* generated NASM version macros */
 
 #ifndef NULL
 #define NULL 0
@@ -33,6 +32,16 @@
 #define FILENAME_MAX 256
 #endif
 
+#ifndef PREFIX_MAX
+#define PREFIX_MAX 10
+#endif
+
+#ifndef POSTFIX_MAX
+#define POSTFIX_MAX 10
+#endif
+
+
+
 /*
  * Name pollution problems: <time.h> on Digital UNIX pulls in some
  * strange hardware header file which sees fit to define R_SP. We
@@ -57,36 +66,38 @@ struct ofmt;
 /*
  * An error reporting function should look like this.
  */
-typedef void (*efunc) (int severity, char *fmt, ...);
+typedef void (*efunc) (int severity, const char *fmt, ...);
 
 /*
  * These are the error severity codes which get passed as the first
  * argument to an efunc.
  */
 
-#define ERR_WARNING 0		       /* warn only: no further action */
-#define ERR_NONFATAL 1		       /* terminate assembly after phase */
-#define ERR_FATAL 2		       /* instantly fatal: exit with error */
-#define ERR_PANIC 3		       /* internal error: panic instantly
+#define ERR_DEBUG  	0x00000008	/* put out debugging message */
+#define ERR_WARNING	0x00000000	/* warn only: no further action */
+#define ERR_NONFATAL	0x00000001	/* terminate assembly after phase */
+#define ERR_FATAL	0x00000002	/* instantly fatal: exit with error */
+#define ERR_PANIC	0x00000003	/* internal error: panic instantly
 					* and dump core for reference */
-#define ERR_MASK 0x0F		       /* mask off the above codes */
-#define ERR_NOFILE 0x10		       /* don't give source file name/line */
-#define ERR_USAGE 0x20		       /* print a usage message */
-#define ERR_OFFBY1 0x40		       /* report error as being on the line 
-					* we're just _about_ to read, not
-					* the one we've just read */
-#define ERR_PASS1 0x80		       /* only print this error on pass one */
+#define ERR_MASK	0x0000000F	/* mask off the above codes */
+#define ERR_NOFILE	0x00000010	/* don't give source file name/line */
+#define ERR_USAGE	0x00000020	/* print a usage message */
+#define ERR_PASS1	0x00000040	/* only print this error on pass one */
 
 /*
  * These codes define specific types of suppressible warning.
  */
-#define ERR_WARN_MNP  0x0100	       /* macro-num-parameters warning */
-#define ERR_WARN_OL   0x0200	       /* orphan label (no colon, and
-					* alone on line) */
-#define ERR_WARN_NOV  0x0300	       /* numeric overflow */
-#define ERR_WARN_MASK 0xFF00	       /* the mask for this feature */
+
+#define ERR_WARN_MASK	0x0000FF00	/* the mask for this feature */
 #define ERR_WARN_SHR  8		       /* how far to shift right */
-#define ERR_WARN_MAX  3		       /* the highest numbered one */
+
+#define ERR_WARN_MNP	0x00000100	/* macro-num-parameters warning */
+#define ERR_WARN_MSR	0x00000200	/* macro self-reference */
+#define ERR_WARN_OL	0x00000300	/* orphan label (no colon, and
+					* alone on line) */
+#define ERR_WARN_NOV	0x00000400	/* numeric overflow */
+#define ERR_WARN_GNUELF	0x00000500      /* using GNU ELF extensions */
+#define ERR_WARN_MAX	5		/* the highest numbered one */
 
 /*
  * -----------------------
@@ -137,7 +148,7 @@ typedef struct {
      * work with when doing things like uplevel(LIST_TIMES) or
      * uplevel(LIST_INCBIN).
      */
-    void (*output) (long, void *, unsigned long);
+    void (*output) (long, const void *, unsigned long);
 
     /*
      * Called to send a text line to the listing generator. The
@@ -205,6 +216,12 @@ enum {				       /* token types, other than chars */
     TOKEN_FLOAT			       /* floating-point constant */
 };
 
+typedef struct {
+    long segment;
+    long offset;
+    int  known;
+} loc_t;
+
 /*
  * Expression-evaluator datatype. Expressions, within the
  * evaluator, are stored as an array of these beasts, terminated by
@@ -247,8 +264,8 @@ struct eval_hints {
  * defined before use", whereas if `critical' is 2, the error will
  * be "symbol undefined".
  *
- * If `critical' has bit 4 set (in addition to its main value: 0x11
- * and 0x12 correspond to 1 and 2) then an extended expression
+ * If `critical' has bit 8 set (in addition to its main value: 0x101
+ * and 0x102 correspond to 1 and 2) then an extended expression
  * syntax is recognised, in which relational operators such as =, <
  * and >= are accepted, as well as low-precedence logical operators
  * &&, ^^ and ||.
@@ -256,17 +273,11 @@ struct eval_hints {
  * If `hints' is non-NULL, it gets filled in with some hints as to
  * the base register in complex effective addresses.
  */
+#define CRITICAL 0x100
 typedef expr *(*evalfunc) (scanner sc, void *scprivate, struct tokenval *tv,
 			   int *fwref, int critical, efunc error,
 			   struct eval_hints *hints);
 
-/*
- * There's also an auxiliary routine through which the evaluator
- * needs to hear about the value of $ and the label (if any)
- * defined on the current line.
- */
-typedef void (*evalinfofunc) (char *labelname, long segment, long offset);
-
 /*
  * Special values for expr->type. ASSUMPTION MADE HERE: the number
  * of distinct register names (i.e. possible "type" fields for an
@@ -301,7 +312,7 @@ typedef struct {
     /*
      * Called at the end of a pass.
      */
-    void (*cleanup) (void);
+    void (*cleanup) (int);
 } Preproc;
 
 /*
@@ -311,11 +322,13 @@ typedef struct {
  * ----------------------------------------------------------------
  */
 
-/* isidstart matches any character that may start an identifier, and isidchar
+/*
+ * isidstart matches any character that may start an identifier, and isidchar
  * matches any character that may appear at places other than the start of an
  * identifier. E.g. a period may only appear at the start of an identifier
  * (for local labels), whereas a number may appear anywhere *but* at the
- * start. */
+ * start. 
+ */
 
 #define isidstart(c) ( isalpha(c) || (c)=='_' || (c)=='.' || (c)=='?' \
                                   || (c)=='@' )
@@ -370,6 +383,7 @@ enum {
 
 #define TO        0x00000100L          /* reverse effect in FADD, FSUB &c */
 #define COLON     0x00000200L	       /* operand is followed by a colon */
+#define STRICT    0x00000400L	       /* do not optimize this operand */
 
 /* type of operand: memory reference, register, etc. */
 #define MEMORY    0x00204000L
@@ -382,6 +396,7 @@ enum {
 #define REG16     0x00201002L
 #define REG32     0x00201004L
 #define MMXREG    0x00201008L	       /* MMX registers */
+#define XMMREG    0x00201010L          /* XMM Katmai reg */
 #define FPUREG    0x01000000L	       /* floating point stack registers */
 #define FPU0      0x01000800L	       /* FPU stack register zero */
 
@@ -400,9 +415,9 @@ enum {
 #define REG_CS    0x01081002L	       /* CS */
 #define REG_DESS  0x02081002L	       /* DS, ES, SS (non-CS 86 registers) */
 #define REG_FSGS  0x04081002L	       /* FS, GS (386 extended registers) */
+#define REG_SEG67 0x08081002L          /* Non-implemented segment registers */
 #define REG_CDT   0x00101004L	       /* CRn, DRn and TRn */
 #define REG_CREG  0x08101004L	       /* CRn */
-#define REG_CR4   0x08101404L	       /* CR4 (Pentium only) */
 #define REG_DREG  0x10101004L	       /* DRn */
 #define REG_TREG  0x20101004L	       /* TRn */
 
@@ -412,73 +427,11 @@ enum {
 /* special type of immediate operand */
 #define ONENESS   0x00800000L          /* so UNITY == IMMEDIATE | ONENESS */
 #define UNITY     0x00802000L	       /* for shift/rotate instructions */
-
-/*
- * Next, the codes returned from the parser, for registers and
- * instructions.
- */
-
-enum {				       /* register names */
-    R_AH = EXPR_REG_START, R_AL, R_AX, R_BH, R_BL, R_BP, R_BX, R_CH,
-    R_CL, R_CR0, R_CR2, R_CR3, R_CR4, R_CS, R_CX, R_DH, R_DI, R_DL,
-    R_DR0, R_DR1, R_DR2, R_DR3, R_DR6, R_DR7, R_DS, R_DX, R_EAX,
-    R_EBP, R_EBX, R_ECX, R_EDI, R_EDX, R_ES, R_ESI, R_ESP, R_FS,
-    R_GS, R_MM0, R_MM1, R_MM2, R_MM3, R_MM4, R_MM5, R_MM6, R_MM7,
-    R_SI, R_SP, R_SS, R_ST0, R_ST1, R_ST2, R_ST3, R_ST4, R_ST5,
-    R_ST6, R_ST7, R_TR3, R_TR4, R_TR5, R_TR6, R_TR7, REG_ENUM_LIMIT
-};
-
-enum {				       /* instruction names */
-    I_AAA, I_AAD, I_AAM, I_AAS, I_ADC, I_ADD, I_AND, I_ARPL,
-    I_BOUND, I_BSF, I_BSR, I_BSWAP, I_BT, I_BTC, I_BTR, I_BTS,
-    I_CALL, I_CBW, I_CDQ, I_CLC, I_CLD, I_CLI, I_CLTS, I_CMC, I_CMP,
-    I_CMPSB, I_CMPSD, I_CMPSW, I_CMPXCHG, I_CMPXCHG486, I_CMPXCHG8B,
-    I_CPUID, I_CWD, I_CWDE, I_DAA, I_DAS, I_DB, I_DD, I_DEC, I_DIV,
-    I_DQ, I_DT, I_DW, I_EMMS, I_ENTER, I_EQU, I_F2XM1, I_FABS,
-    I_FADD, I_FADDP, I_FBLD, I_FBSTP, I_FCHS, I_FCLEX, I_FCMOVB,
-    I_FCMOVBE, I_FCMOVE, I_FCMOVNB, I_FCMOVNBE, I_FCMOVNE,
-    I_FCMOVNU, I_FCMOVU, I_FCOM, I_FCOMI, I_FCOMIP, I_FCOMP,
-    I_FCOMPP, I_FCOS, I_FDECSTP, I_FDISI, I_FDIV, I_FDIVP, I_FDIVR,
-    I_FDIVRP, I_FENI, I_FFREE, I_FIADD, I_FICOM, I_FICOMP, I_FIDIV,
-    I_FIDIVR, I_FILD, I_FIMUL, I_FINCSTP, I_FINIT, I_FIST, I_FISTP,
-    I_FISUB, I_FISUBR, I_FLD, I_FLD1, I_FLDCW, I_FLDENV, I_FLDL2E,
-    I_FLDL2T, I_FLDLG2, I_FLDLN2, I_FLDPI, I_FLDZ, I_FMUL, I_FMULP,
-    I_FNCLEX, I_FNDISI, I_FNENI, I_FNINIT, I_FNOP, I_FNSAVE,
-    I_FNSTCW, I_FNSTENV, I_FNSTSW, I_FPATAN, I_FPREM, I_FPREM1,
-    I_FPTAN, I_FRNDINT, I_FRSTOR, I_FSAVE, I_FSCALE, I_FSETPM,
-    I_FSIN, I_FSINCOS, I_FSQRT, I_FST, I_FSTCW, I_FSTENV, I_FSTP,
-    I_FSTSW, I_FSUB, I_FSUBP, I_FSUBR, I_FSUBRP, I_FTST, I_FUCOM,
-    I_FUCOMI, I_FUCOMIP, I_FUCOMP, I_FUCOMPP, I_FXAM, I_FXCH,
-    I_FXTRACT, I_FYL2X, I_FYL2XP1, I_HLT, I_IBTS, I_ICEBP, I_IDIV,
-    I_IMUL, I_IN, I_INC, I_INCBIN, I_INSB, I_INSD, I_INSW, I_INT,
-    I_INT1, I_INT01, I_INT3, I_INTO, I_INVD, I_INVLPG, I_IRET,
-    I_IRETD, I_IRETW, I_JCXZ, I_JECXZ, I_JMP, I_LAHF, I_LAR, I_LDS,
-    I_LEA, I_LEAVE, I_LES, I_LFS, I_LGDT, I_LGS, I_LIDT, I_LLDT,
-    I_LMSW, I_LOADALL, I_LOADALL286, I_LODSB, I_LODSD, I_LODSW,
-    I_LOOP, I_LOOPE, I_LOOPNE, I_LOOPNZ, I_LOOPZ, I_LSL, I_LSS,
-    I_LTR, I_MOV, I_MOVD, I_MOVQ, I_MOVSB, I_MOVSD, I_MOVSW,
-    I_MOVSX, I_MOVZX, I_MUL, I_NEG, I_NOP, I_NOT, I_OR, I_OUT,
-    I_OUTSB, I_OUTSD, I_OUTSW, I_PACKSSDW, I_PACKSSWB, I_PACKUSWB,
-    I_PADDB, I_PADDD, I_PADDSB, I_PADDSIW, I_PADDSW, I_PADDUSB,
-    I_PADDUSW, I_PADDW, I_PAND, I_PANDN, I_PAVEB, I_PCMPEQB,
-    I_PCMPEQD, I_PCMPEQW, I_PCMPGTB, I_PCMPGTD, I_PCMPGTW,
-    I_PDISTIB, I_PMACHRIW, I_PMADDWD, I_PMAGW, I_PMULHRW,
-    I_PMULHRIW, I_PMULHW, I_PMULLW, I_PMVGEZB, I_PMVLZB, I_PMVNZB,
-    I_PMVZB, I_POP, I_POPA, I_POPAD, I_POPAW, I_POPF, I_POPFD,
-    I_POPFW, I_POR, I_PSLLD, I_PSLLQ, I_PSLLW, I_PSRAD, I_PSRAW,
-    I_PSRLD, I_PSRLQ, I_PSRLW, I_PSUBB, I_PSUBD, I_PSUBSB,
-    I_PSUBSIW, I_PSUBSW, I_PSUBUSB, I_PSUBUSW, I_PSUBW, I_PUNPCKHBW,
-    I_PUNPCKHDQ, I_PUNPCKHWD, I_PUNPCKLBW, I_PUNPCKLDQ, I_PUNPCKLWD,
-    I_PUSH, I_PUSHA, I_PUSHAD, I_PUSHAW, I_PUSHF, I_PUSHFD,
-    I_PUSHFW, I_PXOR, I_RCL, I_RCR, I_RDMSR, I_RDPMC, I_RDTSC,
-    I_RESB, I_RESD, I_RESQ, I_REST, I_RESW, I_RET, I_RETF, I_RETN,
-    I_ROL, I_ROR, I_RSM, I_SAHF, I_SAL, I_SALC, I_SAR, I_SBB,
-    I_SCASB, I_SCASD, I_SCASW, I_SGDT, I_SHL, I_SHLD, I_SHR, I_SHRD,
-    I_SIDT, I_SLDT, I_SMI, I_SMSW, I_STC, I_STD, I_STI, I_STOSB,
-    I_STOSD, I_STOSW, I_STR, I_SUB, I_TEST, I_UMOV, I_VERR, I_VERW,
-    I_WAIT, I_WBINVD, I_WRMSR, I_XADD, I_XBTS, I_XCHG, I_XLATB,
-    I_XOR, I_CMOVcc, I_Jcc, I_SETcc
-};
+#define BYTENESS  0x40000000L          /* so SBYTE == IMMEDIATE | BYTENESS */
+#define SBYTE 	  0x40002000L	       /* for op r16/32,immediate instrs. */
+		
+/* Register names automatically generated from regs.dat */
+#include "regs.h"
 
 enum {				       /* condition code names */
     C_A, C_AE, C_B, C_BE, C_C, C_E, C_G, C_GE, C_L, C_LE, C_NA, C_NAE,
@@ -522,8 +475,12 @@ typedef struct {		       /* operand to an instruction */
     long offset;		       /* any immediate number */
     long wrt;			       /* segment base it's relative to */
     int eaflags;		       /* special EA flags */
+    int opflags;		       /* see OPFLAG_* defines below */
 } operand;
 
+#define OPFLAG_FORWARD		1      /* operand is a forward reference */
+#define OPFLAG_EXTERN		2      /* operand is an external reference */
+
 typedef struct extop {		       /* extended operand */
     struct extop *next;		       /* linked list */
     long type;			       /* defined above */
@@ -542,13 +499,16 @@ typedef struct {		       /* an instruction itself */
     int nprefix;		       /* number of entries in above */
     int opcode;			       /* the opcode - not just the string */
     int condition;		       /* the condition code, if Jcc/SETcc */
-    int operands;		       /* how many operands? 0-3 */
+    int operands;		       /* how many operands? 0-3 
+                                        * (more if db et al) */
     operand oprs[3];	   	       /* the operands, defined as above */
     extop *eops;		       /* extended operands */
+    int eops_float;                    /* true if DD and floating */
     long times;			       /* repeat count (TIMES prefix) */
     int forw_ref;		       /* is there a forward reference? */
 } insn;
 
+enum geninfo { GI_SWITCH };
 /*
  * ------------------------------------------------------------
  * The data structure defining an output format driver, and the
@@ -561,12 +521,33 @@ struct ofmt {
      * This is a short (one-liner) description of the type of
      * output generated by the driver.
      */
-    char *fullname;
+    const char *fullname;
 
     /*
      * This is a single keyword used to select the driver.
      */
-    char *shortname;
+    const char *shortname;
+
+    /*
+     * this is reserved for out module specific help.
+     * It is set to NULL in all the out modules but is not implemented
+     * in the main program
+     */
+    const char *helpstring;
+
+    /*
+     * this is a pointer to the first element of the debug information
+     */
+    struct dfmt **debug_formats;
+
+    /*
+     * and a pointer to the element that is being used
+     * note: this is set to the default at compile time and changed if the
+     * -F option is selected.  If developing a set of new debug formats for
+     * an output format, be sure to set this to whatever default you want
+     *
+     */
+    struct dfmt *current_dfmt;
 
     /*
      * This, if non-NULL, is a NULL-terminated list of `char *'s
@@ -575,7 +556,7 @@ struct ofmt {
      * and user-level equivalents for any format-specific
      * directives).
      */
-    char **stdmac;
+    const char **stdmac;
 
     /*
      * This procedure is called at the start of an output session.
@@ -586,6 +567,15 @@ struct ofmt {
      */
     void (*init) (FILE *fp, efunc error, ldfunc ldef, evalfunc eval);
 
+    /*
+     * This procedure is called to pass generic information to the
+     * object file.  The first parameter gives the information type
+     * (currently only command line switches)
+     * and the second parameter gives the value.  This function returns
+     * 1 if recognized, 0 if unrecognized
+     */
+    int (*setinfo)(enum geninfo type, char **string);
+
     /*
      * This procedure is called by assemble() to write actual
      * generated code or data to the object file. Typically it
@@ -595,7 +585,7 @@ struct ofmt {
      * The `type' argument specifies the type of output data, and
      * usually the size as well: its contents are described below.
      */
-    void (*output) (long segto, void *data, unsigned long type,
+    void (*output) (long segto, const void *data, unsigned long type,
 		    long segment, long wrt);
 
     /*
@@ -705,7 +695,7 @@ struct ofmt {
      * One thing the cleanup routine should always do is to close
      * the output file pointer.
      */
-    void (*cleanup) (void);
+    void (*cleanup) (int debuginfo);
 };
 
 /*
@@ -734,6 +724,103 @@ struct ofmt {
 #define OUT_TYPMASK 0xF0000000UL
 #define OUT_SIZMASK 0x0FFFFFFFUL
 
+/*
+ * ------------------------------------------------------------
+ * The data structure defining a debug format driver, and the
+ * interfaces to the functions therein.
+ * ------------------------------------------------------------
+ */
+
+struct dfmt {
+    
+    /*
+     * This is a short (one-liner) description of the type of
+     * output generated by the driver.
+     */
+    const char *fullname;
+
+    /*
+     * This is a single keyword used to select the driver.
+     */
+    const char *shortname;
+
+
+    /*
+     * init - called initially to set up local pointer to object format, 
+     * void pointer to implementation defined data, file pointer (which
+     * probably won't be used, but who knows?), and error function.
+     */
+    void (*init) (struct ofmt * of, void * id, FILE * fp, efunc error);
+
+    /*
+     * linenum - called any time there is output with a change of
+     * line number or file.
+     */
+    void (*linenum) (const char * filename, long linenumber, long segto);
+
+    /*
+     * debug_deflabel - called whenever a label is defined. Parameters
+     * are the same as to 'symdef()' in the output format. This function
+     * would be called before the output format version.
+     */
+
+    void (*debug_deflabel) (char * name, long segment, long offset,
+                            int is_global, char * special);
+    /*
+     * debug_directive - called whenever a DEBUG directive other than 'LINE'
+     * is encountered. 'directive' contains the first parameter to the
+     * DEBUG directive, and params contains the rest. For example,
+     * 'DEBUG VAR _somevar:int' would translate to a call to this
+     * function with 'directive' equal to "VAR" and 'params' equal to 
+     * "_somevar:int".
+     */
+    void (*debug_directive) (const char * directive, const char * params);
+
+    /*
+     * typevalue - called whenever the assembler wishes to register a type
+     * for the last defined label.  This routine MUST detect if a type was
+     * already registered and not re-register it.
+     */
+    void (*debug_typevalue) (long type);
+
+    /*
+     * debug_output - called whenever output is required
+     * 'type' is the type of info required, and this is format-specific
+     */
+    void (*debug_output) (int type, void *param);
+
+    /*
+     * cleanup - called after processing of file is complete
+     */
+    void (*cleanup) (void);
+
+};
+/*
+ * The type definition macros
+ * for debugging
+ *
+ * low 3 bits: reserved
+ * next 5 bits: type
+ * next 24 bits: number of elements for arrays (0 for labels)
+ */
+
+#define TY_UNKNOWN 0x00
+#define TY_LABEL   0x08
+#define TY_BYTE    0x10
+#define TY_WORD    0x18
+#define TY_DWORD   0x20
+#define TY_FLOAT   0x28
+#define TY_QWORD   0x30
+#define TY_TBYTE   0x38
+#define TY_COMMON  0xE0
+#define TY_SEG     0xE8
+#define TY_EXTERN  0xF0
+#define TY_EQU     0xF8
+
+#define TYM_TYPE(x) ((x) & 0xF8)
+#define TYM_ELEMENTS(x) (((x) & 0xFFFFFF00) >> 8)
+
+#define TYS_ELEMENTS(x)  ((x) << 8)
 /*
  * -----
  * Other
@@ -747,4 +834,17 @@ struct ofmt {
 
 #define elements(x)     ( sizeof(x) / sizeof(*(x)) )
 
+extern int tasm_compatible_mode;
+
+/*
+ * This declaration passes the "pass" number to all other modules
+ * "pass0" assumes the values: 0, 0, ..., 0, 1, 2
+ * where 0 = optimizing pass
+ *       1 = pass 1
+ *       2 = pass 2
+ */
+
+extern int pass0;	/* this is globally known */
+extern int optimizing;
+
 #endif
